/*
 * Copyright (C) 2009-2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.reverb.lang;

import java.io.File;
import java.io.IOException;
import java.lang.annotation.Annotation;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.Manifest;
import java.util.jar.Attributes.Name;

/**
 * {@code ClassResolver} resolves classes matched specified condition from its 
 * own search path.
 * As the default, {@code ClassResolver} can resolve classes by the following 
 * conditions: 
 * <ul>
 * <li>By name</li>
 * <li>By superclass</li>
 * <li>By interface</li>
 * <li>By annotation</li>
 * <li>By {@code Matcher}</li>
 * </ul>
 * 
 * @see ClassLoader
 * @see URLClassLoader
 *
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class ClassResolver {
	
	/**
	 * {@code Matcher} indicates whether the specified class is matched to own 
	 * condition or not.
	 *
	 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
	 */
	public interface Matcher {
		
		/**
		 * Indicates whether the specified class is matched to own condition or 
		 * not.
		 * 
		 * @param clazz The class to be indicated.
		 * @return <code>true</code> If the specified class is matched to own 
		 * condition.
		 */
		public boolean matches(Class<?> clazz);
		
	}
	
	private ClassLoader classLoader;
	
	private List<URL> urls = new ArrayList<URL>();
	
	public void setURLs(List<URL> urls) {
		this.urls = urls;
	}

	/**
	 * Constructs a new {@code ClassResolver} instance with the current context 
	 * {@code ClassLoader}.
	 */
	public ClassResolver() {
		this(Thread.currentThread().getContextClassLoader());
	}
	
	/**
	 * Constructs a new {@code ClassResolver} instance with the specified search 
	 * paths and current context {@code ClassLoader}.
	 * 
	 * @param urls The search path.
	 */
	public ClassResolver(URL... urls) {
		this(Thread.currentThread().getContextClassLoader(), urls);
	}
	
	/**
	 * Constructs a new {@code ClassResolver} instance with the specfied 
	 * {@code ClassLoader} and its search paths.
	 * 
	 * @param classLoader The {@code ClassLoader} to load classes.
	 */
	public ClassResolver(ClassLoader classLoader) {
		if (!(classLoader instanceof URLClassLoader)) {
			throw new IllegalArgumentException(
					"'classLoader' must be instance of URLClassLoader");
		}
		
		this.classLoader = classLoader;
		urls = Arrays.asList(URLClassLoader.class.cast(classLoader).getURLs());
	}
	
	/**
	 * Constructs a new {@code ClassResolver} instance with the specified 
	 * {@code ClassLoader} and search paths.
	 * 
	 * @param classLoader The {@code ClassLoader} to load classes.
	 * @param urls The search path.
	 */
	public ClassResolver(ClassLoader classLoader, URL... urls) {
		this.classLoader = classLoader;
		this.urls = Arrays.asList(urls);
	}
	
	/**
	 * Resolves the {@code Set} view of {@code Class} that contains the 
	 * specified name.
	 * 
	 * @param name The class name
	 * @return The {@code Set} view of {@code Class} that contains the specified 
	 * name.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public Set<Class<?>> resolveByName(final String name) throws IOException,
			ClassNotFoundException {
		Matcher matcher = new Matcher() {
			
			public boolean matches(Class<?> clazz) {
				return clazz.getName().contains(name);
			}
			
		};
		
		return resolve(matcher);
	}
	
	/**
	 * Resolves the {@code Set} view of {@code Class} that inherits the 
	 * specified superclass.
	 * 
	 * @param superclass The superclass to be inherited.
	 * @return The {@code Set} view of {@code Class} that inherits the specified 
	 * superclass.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	@SuppressWarnings("unchecked")
	public <T> Set<Class<? extends T>> resolveBySuperclass(
			final Class<T> superclass) throws IOException,
			ClassNotFoundException {
		Matcher matcher = new Matcher() {
			
			public boolean matches(Class<?> clazz) {
				if (clazz == null) {
					return false;
				}
				
				if (clazz.getSuperclass() != null) {
					if (clazz.getSuperclass().equals(superclass)) {
						return true;
					} else {
						return matches(clazz.getSuperclass());
					}
					
				} else {
					return false;
				}
			}
			
		};
		
		Set<Class<? extends T>> classes = new HashSet<Class<? extends T>>();
		
		for (Class<?> subclass : resolve(matcher)) {
			classes.add((Class<? extends T>) subclass);
		}
		
		return classes;
	}
	
	/**
	 * Resolves the {@code Set} view of {@code Class} that implements the 
	 * specified interface.
	 * 
	 * @param interfaceClass The interface to be implemented.
	 * @return The {@code Set} view of {@code Class} that implements the 
	 * specified interface.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	@SuppressWarnings("unchecked")
	public <T> Set<Class<? extends T>> resolveByInterface(
			final Class<T> interfaceClass) throws IOException,
			ClassNotFoundException {
		Matcher matcher = new Matcher() {
			
			public boolean matches(Class<?> clazz) {
				if (clazz == null) {
					return false;
				}
				
				Class<?>[] interfaces = clazz.getInterfaces();
				
				for (Class<?> c : interfaces) {
					if (c.equals(interfaceClass)) {
						return true;
					}
				}
				
				if (clazz.getSuperclass() != null) {
					return matches(clazz.getSuperclass());
				} else {
					return false;
				}
			}
			
		};
		
		Set<Class<? extends T>> classes = new HashSet<Class<? extends T>>();
		
		for (Class<?> implementation : resolve(matcher)) {
			classes.add((Class<? extends T>) implementation);
		}
		
		return classes;
	}
	
	/**
	 * Resolves the {@code Set} view of {@code Class} be annotated by the 
	 * specified annotation.
	 * 
	 * @param annotation The annotation to be annotated by.
	 * @return The {@code Set} view of {@code Class} be annotated by the 
	 * specified annotation.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public Set<Class<?>> resolveByAnnotation(
			final Class<? extends Annotation> annotation) throws IOException,
			ClassNotFoundException {
		Matcher matcher = new Matcher() {
			
			public boolean matches(Class<?> clazz) {
				if (clazz == null) {
					return false;
				}
				
				Annotation[] annotations = clazz.getAnnotations();
				
				for (Annotation a : annotations) {
					if (a.annotationType().equals(annotation)) {
						return true;
					}
				}
				
				return false;
			}
			
		};
		
		return resolve(matcher);
	}
	
	/**
	 * Resolves the {@code Set} view of {@code Class} matches the specified 
	 * {@code Matcher}.
	 * 
	 * @param matcher The {@code Matcher}.
	 * @return The {@code Set} view of {@code Class} matches the specified 
	 * {@code Matcher}.
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public Set<Class<?>> resolve(Matcher matcher) throws IOException,
			ClassNotFoundException {
		Set<Class<?>> classes = new HashSet<Class<?>>();
		
		for (URL url : urls) {
			if (url.toString().endsWith(".jar")) {
				JarFile jarFile = new JarFile(URLDecoder.decode(url.getPath(), "UTF-8"));
				Manifest manifest = jarFile.getManifest();
				
				if (manifest != null) {
					Attributes mainAttributes = manifest.getMainAttributes();
					
					if (mainAttributes != null) {
						String classpath = mainAttributes.getValue(Name.CLASS_PATH);
						
						if (classpath != null) {
							StringTokenizer stringTokenizer = new StringTokenizer(classpath);
							
							while (stringTokenizer.hasMoreTokens()) {
								String token = stringTokenizer.nextToken();
								URL entry = new URL(url, token);
								
								if (urls.contains(entry)) {
									if (entry.toString().endsWith("/")) {
										// System.out.println(entry);
										classes.addAll(getMatchedClasses(matcher, new File(URLDecoder.decode(entry.getPath(), "UTF-8"))));
									} else {
										classes.addAll(getMatchedClasses(matcher, new JarFile(URLDecoder.decode(entry.getPath(), "UTF-8"))));
									}
								}
							}
						}
					}
				}
				
				classes.addAll(getMatchedClasses(matcher, jarFile));
			} else {
				File base = new File(URLDecoder.decode(url.getPath(), "UTF-8"));
				classes.addAll(getMatchedClasses(matcher, base));
			}
		}
		
		return classes;
	}
	
	private Set<Class<?>> getMatchedClasses(Matcher matcher, File base)
			throws ClassNotFoundException {
		List<File> files = getClasses(base);
		Set<Class<?>> classes = new HashSet<Class<?>>();
		
		for (File file : files) {
			String path = file.getPath();
			
			try {
				Class<?> clazz = Class.forName(path.substring(
						base.getPath().length() + 1, path.length() - 6).replace(
						File.separatorChar, '.'), false, classLoader);
				
				if (matcher.matches(clazz)) {
					classes.add(clazz);
				}
				
			} catch (NoClassDefFoundError e) {}
		}
		
		return classes;
	}
	
	private Set<Class<?>> getMatchedClasses(Matcher matcher, JarFile jarFile)
			throws ClassNotFoundException {
		Enumeration<JarEntry> entries = jarFile.entries();
		Set<Class<?>> classes = new HashSet<Class<?>>();
		
		while (entries.hasMoreElements()) {
			JarEntry entry = entries.nextElement();
			String name = entry.getName();
			
			if (name.endsWith(".class")) {
				try {
					Class<?> clazz = Class.forName(name.substring(0,
							name.length() - 6).replace('/', '.'), false,
							classLoader);
					
					if (matcher.matches(clazz)) {
						classes.add(clazz);
					}
					
				} catch (NoClassDefFoundError e) {
				} catch (ClassNotFoundException e) {}
			}
		}
		
		return classes;
	}
	
	private List<File> getClasses(File base) {
		// System.out.println(directory);
		List<File> classes = new ArrayList<File>();
		
		if (base.isDirectory()) {
			File[] files = base.listFiles();
			
			for (File file : files) {
				if (file.isDirectory()) {
					classes.addAll(getClasses(file));
				} else {
					if (file.getPath().endsWith(".class")) {
						classes.add(file);
					}
				}
			}
			
		} else {
			if (base.getPath().endsWith(".class")) {
				classes.add(base);
			}
		}
		
		return classes;
	}

	public ClassLoader getClassLoader() {
		return classLoader;
	}

	public List<URL> getURLs() {
		return urls;
	}
	
}
